def is_iterable(obj:any):
    try:
        iterator=iter(obj)
    except TypeError:
        return False
    else:
        return True
    return True

def get_type(obj:any) -> any:
    """
    Determines the type of the input object.

    Args:
        obj: The object to determine the type of.

    Returns:
        any: The object with the updated type.
    """
    if is_number(obj):
        obj = int(obj)
    if is_float(obj):
        return float(obj)
    elif obj == 'None':
        obj = None
    elif is_str(obj):
        obj = str(obj)
    return obj

def is_instance(obj:any,typ:any) -> bool:
    """
    Checks whether the input object can be represented as a number.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object can be represented as a number, False otherwise.
    """
    boolIt = False
    try:
        boolIt = isinstance(obj, typ)
        return boolIt
    except:
        return boolIt

def is_number(s):
    try:
        float(s)
        return True
    except:
        return False

def is_object(obj:any) -> bool:
    """
    Checks whether the input object is of type 'object'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'object', False otherwise.
    """
    return is_instance(obj, object)
def is_str(obj:any) -> bool:
    """
    Checks whether the input object is of type 'str'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'str', False otherwise.
    """
    return is_instance(obj, str)
def is_int(obj:any) -> bool:
    """
    Checks whether the input object is of type 'int'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'int', False otherwise.
    """
    return is_instance(obj, int)
def is_float(obj:any) -> bool:
    """
    Checks whether the input object is of type 'float'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'float', False otherwise.
    """
    return is_instance(obj, float)
def is_bool(obj:any) -> bool:
    """
    Checks whether the input object is of type 'bool'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'bool', False otherwise.
    """
    return is_instance(obj, bool)


def is_list(obj:any) -> bool:
    """
    Checks whether the input object is of type 'list'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'list', False otherwise.
    """
    return is_instance(obj, list)
def is_tuple(obj:any) -> bool:
    """
    Checks whether the input object is of type 'tuple'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'tuple', False otherwise.
    """
    return is_instance(obj, tuple)
def is_set(obj:any) -> bool:
    """
    Checks whether the input object is of type 'set'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'set', False otherwise.
    """
    return is_instance(obj, set)
def is_dict(obj:any) -> bool:
    """
    Checks whether the input object is of type 'dict'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'dict', False otherwise.
    """
    return is_instance(obj, dict)
def is_frozenset(obj:any) -> bool:
    """
    Checks whether the input object is of type 'frozenset'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'frozenset', False otherwise.
    """
    return is_instance(obj, frozenset)
def is_bytearray(obj:any) -> bool:
    """
    Checks whether the input object is of type 'bytearray'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'bytearray', False otherwise.
    """
    return is_instance(obj, bytearray)
def is_bytes(obj:any) -> bool:
    """
    Checks whether the input object is of type 'bytes'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'bytes', False otherwise.
    """
    return is_instance(obj, bytes)
def is_memoryview(obj:any) -> bool:
    """
    Checks whether the input object is of type 'memoryview'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'memoryview', False otherwise.
    """
    return is_instance(obj, memoryview)
def is_range(obj:any) -> bool:
    """
    Checks whether the input object is

 of type 'range'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'range', False otherwise.
    """
    return is_instance(obj, range)
def is_enumerate(obj:any) -> bool:
    """
    Checks whether the input object is of type 'enumerate'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'enumerate', False otherwise.
    """
    return is_instance(obj, enumerate)
def is_zip(obj:any) -> bool:
    """
    Checks whether the input object is of type 'zip'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'zip', False otherwise.
    """
    return is_instance(obj, zip)
def is_filter(obj:any) -> bool:
    """
    Checks whether the input object is of type 'filter'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'filter', False otherwise.
    """
    return is_instance(obj, filter)
def is_map(obj:any) -> bool:
    """
    Checks whether the input object is of type 'map'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'map', False otherwise.
    """
    return is_instance(obj, map)
def is_property(obj:any) -> bool:
    """
    Checks whether the input object is of type 'property'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'property', False otherwise.
    """
    return is_instance(obj, property)


def is_slice(obj:any) -> bool:
    """
    Checks whether the input object is of type 'slice'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'slice', False otherwise.
    """
    return is_instance(obj, slice)


def is_super(obj:any) -> bool:
    """
    Checks whether the input object is of type 'super'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'super', False otherwise.
    """
    return is_instance(obj, super)


def is_type(obj:any) -> bool:
    """
    Checks whether the input object is of type 'type'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'type', False otherwise.
    """
    return is_instance(obj, type)


def is_Exception(obj:any) -> bool:
    """
    Checks whether the input object is of type 'Exception'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'Exception', False otherwise.
    """
    return is_instance(obj, Exception)


def is_none(obj:any) -> bool:
    """
    Checks whether the input object is of type 'None'.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'None', False otherwise.
    """
    if type(obj) is None:
        return True
    else:
        return False



def is_dict_or_convertable(obj:any) -> bool:
    """
    Checks whether the input object is of type 'dict' or can be converted to a dictionary.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object is of type 'dict' or can be converted to a dictionary, False otherwise.
    """
    if is_dict(obj):
        return True
    if is_str_convertible_dict(obj):
        return True
    return False
def is_str_convertible_dict(obj:any) -> bool:
    """
    Checks whether the input object is a string that can be converted to a dict.

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object can be converted to a dict, False otherwise.
    """
    import json

    if is_instance(obj, str):
        try:
            json.loads(obj)
            return True
        except json.JSONDecodeError:
            return False

    return False
def is_any_instance(value):
    for each in [dict, list, int, float]:
        if is_instance(value, each):
            return True
def det_bool_F(obj: (tuple or list or bool) = False):
    """
    Determines if the given object is a boolean False value.

    Args:
        obj (tuple or list or bool): The object to determine the boolean False value.

    Returns:
        bool: True if the object is a boolean False value, False otherwise.
    """
    if is_instance(obj, bool):
        return obj
    return all(obj)
def det_bool_T(obj: (tuple or list or bool) = False):
    """
    Determines if the given object is a boolean True value.

    Args:
        obj (tuple or list or bool): The object to determine the boolean True value.

    Returns:
        bool: True if the object is a boolean True value, False otherwise.
    """
    if is_instance(obj, bool):
        return obj 
    return any(obj)
def T_or_F_obj_eq(event: any = '', obj: any = ''):
    """
    Compares two objects and returns True if they are equal, False otherwise.

    Args:
        event (any): The first object to compare.
        obj (any): The second object to compare.

    Returns:
        bool: True if the objects are equal, False otherwise.
    """
    return True if event == obj else False
def ensure_integer(page_value:any, default_value:int):
    """
    Ensures the given value is an integer. If not, it tries to extract 
    the numeric part of the value. If still unsuccessful, it defaults 
    to the given default value.

    Parameters:
    - page_value (str|int|any): The value to ensure as integer. 
                                Non-numeric characters are stripped if necessary.
    - default_value (int): The default value to return if conversion 
                           to integer is unsuccessful.

    Returns:
    - int: The ensured integer value.
    """
    # Check if page_value is already a number
    if not is_number(page_value):
        # Convert to string in case it's not already
        page_value = str(page_value)
        
        # Remove non-numeric characters from the beginning
        while len(page_value) > 0 and page_value[0] not in '0123456789'.split(','):
            page_value = page_value[1:]

        # Remove non-numeric characters from the end
        while len(page_value) > 0 and page_value[-1] not in '0123456789'.split(','):
            page_value = page_value[:-1]

    # If page_value is empty or still not a number, use the default value
    if len(page_value) == 0 or not is_number(page_value):
        return default_value

    # Convert page_value to an integer and return
    return int(page_value)
def if_default_return_obj(obj:any,default:any=None,default_compare:any=None):
    if default == default_compare:
        return obj
    return default


            



from .eat_utils import eatAll
def clean_spaces(obj: str) -> str:
    """
    Removes leading spaces and tabs from a string.

    Args:
        obj (str): The input string.

    Returns:
        str: The string with leading spaces and tabs removed.
    """
    if len(obj) == 0:
        return obj
    while obj[0] in [' ', '\t']:
        obj = obj[1:]
    return obj
def clean_line(line):
    return eatAll(line,[' ','','\t','\n'])
def url_join(*paths):
    final_url = os.path.join(*paths)
    for i,path in enumerate(paths):
        if i == 0:
            final_path = path  # Note: Fixed bug; original code had `final_path = paths`
        else:
            final_path = eatOuter(final_path, '/')
            path = eatInner(path, '/')
            final_path = f"{final_path}/{path}"
    return final_path



from .imports import *
from .module_imports import *
def get_pass_from_key(key=None,env_path=None):
    if key:
        return get_env_value(key=key,path=env_path)
def get_password(password=None,key=None,env_path=None):
    password = password or get_pass_from_key(key=key,env_path=env_path)
    return password

def get_print_sudo_cmd(
        cmd: str,
        password=None,
        key=None,
        env_path=None
    ):
    password = get_password(password=password,key=key,env_path=env_path)
    if password != None:
        
        cmd = get_password_cmd(password=password,cmd=cmd)
    return cmd
def get_password_cmd(password:str,cmd:str):
    sudo_cmd = get_sudo_cmd(cmd)
    password_sudo_cmd = get_raw_password_sudo_cmd(password=password,sudo_cmd=sudo_cmd)
    return password_sudo_cmd
def get_sudo_cmd(cmd: str):
    return f"sudo -S -k {cmd}"
def get_raw_password_sudo_cmd(password:str,sudo_cmd:str):
    return f"printf %s {shlex.quote(password)} | {sudo_cmd}"
def get_remote_bash(
            cmd: str,
            cwd: str | None = None
        ):
    return f"bash -lc {shlex.quote((f'cd {shlex.quote(cwd)} && {cmd}') if cwd else cmd)}"
def get_remote_ssh(
            user_at_host: str=None,
            remote:str=None
        ):
    return f"ssh {shlex.quote(user_at_host)} {shlex.quote(remote)}"
def get_remote_cmd(
            cmd: str,
            user_at_host: str,
            cwd: str | None = None,
            password=None,
            key=None,
            env_path=None
            
        ):
    cmd = get_print_sudo_cmd(
            cmd=cmd,
            password=password,
            key=key,
            env_path=env_path
        )
    remote = get_remote_bash(
        cmd=cmd,
        cwd=cwd
        )
    full = get_remote_ssh(
        user_at_host=user_at_host,
        remote=remote
        )
    return full


def execute_cmd(
        *args,
        outfile=None,
        **kwargs
    ) -> str:
    proc = subprocess.run(*args, **kwargs)
    output = (proc.stdout or "") + (proc.stderr or "")
    if outfile:
        try:
            with open(outfile, "w", encoding="utf-8", errors="ignore") as f:
                f.write(output)
        except Exception:
            pass
    return output

def run_local_cmd(
        cmd: str,
        cwd: str | None = None,
        outfile: Optional[str] = None,
        shell=True,
        text=True,
        capture_output=True,
        user_at_host: str=None,
        password=None,
        key=None,
        env_path=None
    ) -> str:
    cmd = get_print_sudo_cmd(
            cmd=cmd,
            password=password,
            key=key,
            env_path=env_path
        )
    return execute_cmd(
            cmd,
            outfile=outfile,
            shell=shell,
            cwd=cwd,
            text=text,
            capture_output=capture_output
        )

def run_remote_cmd(
        user_at_host: str,
        cmd: str,
        cwd: str | None = None,
        outfile: Optional[str] = None,
        shell=True,
        text=True,
        capture_output=True,
        password=None,
        key=None,
        env_path=None
    ) -> str:
    """
    Run on remote via SSH; capture stdout+stderr locally; write to local outfile.
    NOTE: we do *not* try to write the file on the remote to avoid later scp.
    """
    cmd = get_print_sudo_cmd(
            cmd=cmd,
            password=password,
            key=key,
            env_path=env_path
        )
    # wrap in bash -lc for PATH/profile + allow 'cd && ...'
    cmd = get_remote_cmd(
        cmd=cmd,
        user_at_host=user_at_host,
        cwd=cwd
        )
    return execute_cmd(
            cmd,
            outfile=outfile,
            shell=shell,
            text=text,
            capture_output=capture_output
        )

def run_cmd(
        cmd: str=None,
        cwd: str | None = None,
        outfile: Optional[str] = None,
        shell=True,
        text=True,
        capture_output=True,
        user_at_host: str=None,
        password=None,
        key=None,
        env_path=None
    ) -> str:

    if user_at_host:
        return run_ssh_cmd(
                user_at_host=user_at_host,
                cmd=cmd,
                cwd=cwd,
                outfile=outfile,
                shell=shell,
                text=text,
                capture_output=capture_output,
                password=password,
                key=key,
                env_path=env_path
            )
    return run_local_cmd(
            cmd=cmd,
            cwd=cwd,
            outfile=outfile,
            shell=shell,
            text=text,
            capture_output=capture_output,
            password=password,
            key=key,
            env_path=env_path
        )
run_ssh_cmd = run_remote_cmd
remote_cmd = run_remote_cmd
ssh_cmd = run_remote_cmd

local_cmd = run_local_cmd


run_any_cmd = run_cmd
any_cmd = run_cmd
cmd_run = run_cmd



from .imports import *
def iter_values(obj: JSONLike, path: PathType = ()) -> Iterator[Tuple[PathType, Any]]:
    """
    Depth-first walk of nested dict/list/tuple/set. Yields (path, value) for every leaf.
    Path items are str (dict key) or int (list index).
    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            yield from iter_values(v, path + (k,))
    elif isinstance(obj, (list, tuple)):
        for i, v in enumerate(obj):
            yield from iter_values(v, path + (i,))
    elif isinstance(obj, set):
        for v in obj:
            # sets are unordered/no index – use a pseudo-key
            yield from iter_values(v, path + ('<setitem>',))
    else:
        # leaf (scalar)
        yield path, obj

def _mk_predicate(
    terms: Iterable[str] | None = None,
    *,
    case_insensitive: bool = True,
    substring: bool = True,
    regex: bool = False
) -> Callable[[Any], bool]:
    """
    Build a predicate that checks a scalar value against terms.
    """
    terms = list(terms or [])
    if regex:
        flags = re.IGNORECASE if case_insensitive else 0
        patterns = [re.compile(t, flags) for t in terms]
        def pred(value: Any) -> bool:
            if not isinstance(value, str):
                return False
            return any(p.search(value) for p in patterns)
        return pred

    # string contains (or equals)
    def pred(value: Any) -> bool:
        if not isinstance(value, str):
            return False
        v = value.lower() if case_insensitive else value
        for t in terms:
            t2 = t.lower() if case_insensitive else t
            if (t2 in v) if substring else (t2 == v):
                return True
        return False

    return pred

def search_values(
    obj: JSONLike,
    *,
    terms: Iterable[str],
    case_insensitive: bool = True,
    substring: bool = True,
    regex: bool = False,
) -> List[Tuple[PathType, Any]]:
    """
    Return all (path, value) where value matches any term.
    """
    pred = _mk_predicate(terms, case_insensitive=case_insensitive, substring=substring, regex=regex)
    results: List[Tuple[PathType, Any]] = []
    for p, v in iter_values(obj):
        if pred(v):
            results.append((p, v))
    return results

def any_match(
    obj: JSONLike,
    *,
    terms: Iterable[str],
    **kw
) -> bool:
    """Fast boolean check."""
    pred = _mk_predicate(terms, **kw)
    for _, v in iter_values(obj):
        if pred(v):
            return True
    return False
def get_first_match(obj: JSONLike, *, terms: Iterable[str], **kw) -> Optional[Any]:
    """Return just the first matching value, or None."""
    pred = _mk_predicate(terms, **kw)
    for _, v in iter_values(obj):
        if pred(v):
            return v
    return None

def get_all_match(obj: JSONLike, *, terms: Iterable[str], **kw) -> List[Any]:
    """Return all matching values as a flat list."""
    pred = _mk_predicate(terms, **kw)
    results: List[Any] = []
    for _, v in iter_values(obj):
        if pred(v):
            results.append(v)
    return results



from ..imports import *
from .dot_utils import *
from .sysroot_utils import *
from .extract_utils import *
def find_top_package_dir(p: Path) -> Path | None:
    p = p.resolve()
    if p.is_file():
        p = p.parent
    top = None
    while (p / "__init__.py").exists():
        top = p
        if p.parent == p:
            break
        p = p.parent
    return top
def safe_import(name: str, package: str | None = None, member: str | None = None):
    """
    Wraps importlib.import_module but also resolves relative imports like '..logPaneTab'.
    If `member` is given, returns that attribute from the module.
    """
    mod = importlib.import_module(name, package=package)
    return getattr(mod, member) if member else mod
def unique_module_name(base: str, path: Path) -> str:
    # Make a stable, unique name per file
    digest = hashlib.md5(str(path.resolve()).encode()).hexdigest()[:8]
    stem = path.stem.replace('-', '_')
    return f"_dyn.{base}.{stem}_{digest}"

def import_from_path(module_name: str, file_path: Path):
    spec = importlib.util.spec_from_file_location(module_name, str(file_path))
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load spec for {file_path}")
    mod = importlib.util.module_from_spec(spec)
    # Register before exec to satisfy intra-module relative imports (rare but safe)
    sys.modules[module_name] = mod
    spec.loader.exec_module(mod)
    return mod 
def get_imports(files: List[str],self=None) -> Dict[str, Dict[str, Any]]:
    """
    Discover package context from each file's relative-import requirements,
    add the proper sysroot to sys.path once, and import by dotted name.
    Returns: {modname: {"filepath", "sysroot", "module", "class", "funcs", "selfs"}}
    """
    results: Dict[str, Dict[str, Any]] = {}
    seen_sysroots: set[str] = set()
    files = make_list(files)
    for file in files:
        file_p = Path(file).resolve()

        # 1) Use your logic to find a viable sysroot (top where dots won't exceed)
        sysroot_guess = Path(get_dot_range_sysroot(str(file_p))).resolve()
        # 2) Ensure we import with *package* context: we need the directory ABOVE
        #    the topmost package on sys.path, not the package dir itself.
        top_pkg_dir = find_top_package_dir(sysroot_guess) or find_top_package_dir(file_p)
        if not top_pkg_dir:
            raise RuntimeError(f"No package context found for {file_p}; add __init__.py up the tree.")
        sysroot = top_pkg_dir.parent

        if str(sysroot) not in seen_sysroots:
            ensure_on_path(sysroot)
            seen_sysroots.add(str(sysroot))

        # 3) Compute dotted name from sysroot and import
        dotted = dotted_from(file_p, sysroot)
        try:
            mod = importlib.import_module(dotted)
        except Exception as e:
            # Helpful hint if user ran file directly (no package context)
            if "__package__" in dir() and not __package__:
                raise RuntimeError(
                    f"Import failed for {dotted}. If you ran this file directly, "
                    f"bootstrap package context or run via `python -m ...`."
                ) from e
            raise

        # 4) Collect symbols (you can keep your regex readers if you prefer)
        classes = extract_class(str(file_p))
        funcs   = extract_funcs(str(file_p))
        selfs   = extract_selfs(str(file_p))
        
        # stable result key (avoid collisions)
        key = get_unique_name(get_file_parts(str(file_p))['filename'], results)

        results[key] = {
            "filepath": str(file_p),
            "sysroot": str(sysroot),
            "module": mod,
            "class": classes,
            "funcs": funcs,
            "selfs": selfs,
        }

    return results
def inject_symbols_from_module(
    mod,
    into: dict,
    *,
    expose_functions: bool = True,
    expose_classes: bool = True,
    expose_methods: bool = False,
    self: object | None = None,   # <---- NEW
    ctor_overrides: Mapping[str, Tuple[tuple, dict]] | None = None,
    name_prefix: str = "",
    name_filter: Iterable[str] | None = None,
    overwrite: bool = False
) -> Dict[str, Any]:
    """
    Inject functions/classes from `mod` into `into` (usually globals()).
    If expose_methods=True and a `self` object is passed, bind instance methods
    directly onto `self` instead of globals().
    """
    exported: Dict[str, Any] = {}
    ctor_overrides = ctor_overrides or {}

    def allowed(name: str) -> bool:
        return name_filter is None or name in name_filter

    # 1) functions
    if expose_functions:
        for name, obj in vars(mod).items():
            input(name)
            if not allowed(name):
                continue
            if inspect.isfunction(obj) and obj.__module__ == mod.__name__:
                out_name = f"{name_prefix}{name}"
                if overwrite or out_name not in into:
                    into[out_name] = obj
                    exported[out_name] = obj

    # 2) classes (+ optional bound methods)
    if expose_classes or expose_methods:
        for cls_name, cls in vars(mod).items():
            if not allowed(cls_name):
                continue
            if inspect.isclass(cls) and cls.__module__ == mod.__name__:
                if expose_classes:
                    out_name = f"{name_prefix}{cls_name}"
                    if overwrite or out_name not in into:
                        into[out_name] = cls
                        exported[out_name] = cls

                if expose_methods and self is not None:
                    # instantiate class
                    args, kwargs = ctor_overrides.get(cls_name, ((), {}))
                    try:
                        inst = cls(*args, **kwargs)
                    except Exception:
                        continue

                    for meth_name, meth_obj in vars(cls).items():
                        if meth_name.startswith("__"):
                            continue
                        if inspect.isfunction(meth_obj) or inspect.ismethoddescriptor(meth_obj):
                            try:
                                bound = getattr(inst, meth_name)
                            except Exception:
                                continue
                            if callable(bound):
                                # attach directly to the `self` you passed in
                                if not hasattr(self, meth_name) or overwrite:
                                    setattr(self, meth_name, bound)
                                    exported[f"{cls_name}.{meth_name}"] = bound

    return exported
def inject_from_imports_map(
    imports_map: Dict[str, Dict[str, Any]],
    *,
    into: Optional[dict] = None,      # where to put free funcs/classes (defaults to this module's globals)
    attach_self: Optional[object] = None,  # bind names listed in "selfs" onto this object
    prefix_modules: bool = False,     # add "<module>__" prefix to avoid collisions
    overwrite: bool = False,           # allow overwriting existing names
    self:any=None
) -> Dict[str, Any]:
    """
    Emulates: from functions import *   (plus: bind 'self' methods)
    Returns dict of injected_name -> object (including methods bound to attach_self).
    """
    ns = into if into is not None else globals()
    exported: Dict[str, Any] = {}

    for mod_key, info in imports_map.items():
        mod = info.get("module")
        if mod is None:
            continue

        func_names: List[str]  = info.get("funcs", [])
        self_names: List[str]  = info.get("selfs", [])
        class_names: List[str] = info.get("class", [])

        # 1) bind the "selfs" directly onto self
        if self is not None:
            for name in self_names:
                func = getattr(mod, name, None)
                if not callable(func):
                    continue
                # sanity: ensure first param is literally named 'self'
                try:
                    params = list(inspect.signature(func).parameters.values())
                    if not params or params[0].name != "self":
                        continue
                except Exception:
                    continue

                bound = MethodType(func, self)
                if overwrite or not hasattr(self, name):
                    setattr(self, name, bound)
                    exported[f"{mod_key}.self.{name}"] = bound

        # 2) Inject free functions (exclude the ones we just bound to self to avoid dupes)
        for name in func_names:
            if name in self_names:
                continue
            obj = getattr(mod, name, None)
            if inspect.isfunction(obj) and obj.__module__ == mod.__name__:
                out_name = f"{mod_key}__{name}" if prefix_modules else name
                if overwrite or out_name not in ns:
                    ns[out_name] = obj
                    exported[out_name] = obj

        # 3) Inject classes
        for name in class_names:
            cls = getattr(mod, name, None)
            if inspect.isclass(cls) and cls.__module__ == mod.__name__:
                out_name = f"{mod_key}__{name}" if prefix_modules else name
                if overwrite or out_name not in ns:
                    ns[out_name] = cls
                    exported[out_name] = cls

    # optional: control wildcard export from THIS module
    ns["__all__"] = sorted(set(list(ns.get("__all__", [])) + list(exported.keys())))
    return exported
def ifFunctionsInFiles(root: str | None = None, *, expose_methods: bool = True, self=None) -> Dict[str, Any]:
    here = Path(__file__).resolve().parent
    base = Path(root).resolve() if root else here
    candidates = [base / "functions", base / "functions.py"]
    files: List[str] = []
    for item in candidates:
        if item.exists():
            if item.is_dir():
                files = [str(p) for p in item.rglob("*.py") if p.name != "__init__.py"]
            else:
                files = [str(item)]
            break

    exported_all: Dict[str, Any] = {}
    imports_map = get_imports(files)
    
    exported = inject_from_imports_map(
            imports_map,
            self=self
        )
        
    return exported
def attach_self_functions(
    imports_map: Dict[str, Dict[str, Any]],
    self_obj: object,
    *,
    overwrite: bool = False,
    only: Optional[List[str]] = None,     # whitelist of method names to attach
    prefix_modules: bool = False          # attach as core_utils__on_run_code instead of on_run_code
) -> Dict[str, Any]:
    """
    For each module in imports_map, bind names listed in 'selfs' (top-level defs whose
    first param is 'self') directly onto `self_obj`.

    Returns {attached_name: bound_method}.
    """
    attached: Dict[str, Any] = {}

    for mod_key, info in imports_map.items():
        mod = info.get("module")
        if not mod:
            continue

        self_names: List[str] = info.get("selfs", [])
        for name in self_names:
            if only is not None and name not in only:
                continue

            func = getattr(mod, name, None)
            if not callable(func):
                continue

            # sanity check: first param literally named 'self'
            try:
                params = list(inspect.signature(func).parameters.values())
                if not params or params[0].name != "self":
                    continue
            except Exception:
                continue

            bound = MethodType(func, self_obj)
            out_name = f"{mod_key}__{name}" if prefix_modules else name

            if overwrite or not hasattr(self_obj, out_name):
                setattr(self_obj, out_name, bound)
                attached[out_name] = bound

    return attached
def initFuncs(self, mode=None):
    """
    Load functions in either dev (dynamic) or prod (static) mode.
    """
    mode = mode or "dev"
    from abstract_utilities import get_logFile
    logger = get_logFile(__name__)

    try:
        base_pkg = self.__class__.__module__.rsplit('.', 1)[0]

        if mode == "prod":
            # All imports come from the auto-generated import script
            imports_mod = importlib.import_module(f"{base_pkg}.imports.funcs_imports")

            # Attach all exports
            for name in getattr(imports_mod, "__all__", []):
                func = getattr(imports_mod, name)
                setattr(self, name, func.__get__(self))

            return self

        # --- DEV MODE: your existing dynamic loader ---
        import inspect, pkgutil, importlib
        pkg = importlib.import_module(base_pkg + ".functions")

        existing = set(self.__dict__) | set(dir(self))

        for _, module_name, _ in pkgutil.iter_modules(pkg.__path__):
            mod = importlib.import_module(f"{pkg.__name__}.{module_name}")

            for name, obj in inspect.getmembers(mod, inspect.isfunction):
                if name not in existing:
                    setattr(self, name, obj.__get__(self))

    except Exception as e:
        logger.error(f"initFuncs({mode}) error: {e}")

    return self



#!/usr/bin/env python3
from .imports import *
DEFAULT_FILE_NAME = '.env'
DEFAULT_KEY = 'MY_PASSWORD'
def find_and_read_env_file(key:str=DEFAULT_KEY,file_name:str=DEFAULT_FILE_NAME,  start_path:str=None):
    """
    Search for an environment file and read a specific key from it.

    Args:
        file_name (str): Name of the .env file to be searched. Defaults to '.env'.
        key (str): Key to be retrieved from the .env file. Defaults to 'MY_PASSWORD'.
        start_path (str): Directory path to start the search from. If None, search starts from current directory. 

    Returns:
        str: The value corresponding to the key if found, otherwise None.
    """
    # Set the default start_path to the current directory if it's None
    directories = [start_path, os.getcwd(), get_home_folder(), simple_path_join(get_home_folder(),'.envy_all'),simple_path_join(get_home_folder(),'envy_all')]
    if start_path in [None, os.getcwd()]:
        directories = directories[1:]

    # Try to find the file in the start_path
    for k in range(0,len(directories)):
        env_path = check_env_file(path=directories[k],file_name=file_name)
        if not is_bool(env_path):
            value = search_for_env_key(path=env_path,key=key)
            if value != None:
                return value

def search_for_env_key(key:str,path:str):
    """
    Search for a specific key in a .env file.

    Args:
        path (str): The path to the .env file.
        key (str): The key to search for in the .env file.

    Returns:
        str: The value of the key if found, otherwise None.
    """
    with open(path, "r") as f:
        for line in f:
            eq_split = safe_split(line,['=',0])
            # If the line contains the key, return the value after stripping extra characters
            if line_contains(string=eq_split, compare=key):
                return eatAll(line[len(eq_split):],[' ','','=']).strip()

def check_env_file(path:str,file_name:str=DEFAULT_FILE_NAME):
    """
    Check if the environment file exists in a specified path.

    Args:
        path (str): The path to check for the .env file.
        file_name (str): The name of the .env file. Defaults to '.env'.

    Returns:
        str: The path of the .env file if it exists, otherwise False.
    """
    path = if_not_last_child_join(path=path, child=DEFAULT_FILE_NAME)
    # Return the path if file exists, otherwise return False
    if is_file(path):
        return path
    return False

def safe_env_load(path:str=None):
    """
    Safely load the .env file if it exists at a specified path.

    Args:
        path (str): The path to load the .env file from. If None, no operation is performed. 

    Returns:
        bool: True if the .env file is successfully loaded, otherwise False.
    """
    if path == None:
        return False
    if is_file(path):
        if str(safe_split(path,[get_slash(),-1]))[0] == '.':
            load_dotenv(path)
            return True
    return False

def get_env_value(key:str=DEFAULT_KEY,path:str=None,file_name:str=DEFAULT_FILE_NAME):
    """
    Retrieves the value of the specified environment variable.

    Args:
        path (str): The path to the environment file. Defaults to None.
        file_name (str): The name of the environment file. Defaults to '.env'.
        key (str): The key to search for in the .env file. Defaults to 'MY_PASSWORD'.

    Returns:
        str: The value of the environment variable if found, otherwise None.
    """
    if safe_env_load(path):
        return os.getenv(key)
    return find_and_read_env_file(file_name=file_name, key=key, start_path=os.getcwd())



from __future__ import annotations

import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path
import sys
import os

PACKAGE_NAME = "abstract_utilities"

# ─────────────────────────────────────────────────────────────
# Formatting
# ─────────────────────────────────────────────────────────────

LOG_FORMAT = (
    "[%(asctime)s] "
    "%(levelname)-8s "
    "%(name)s:%(lineno)d | "
    "%(message)s "
    "[target=%(target_file)s:%(target_line)s]"
)

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class SafeFormatter(logging.Formatter):
    """Formatter that tolerates missing `extra` fields."""
    def format(self, record: logging.LogRecord) -> str:
        record.target_file = getattr(record, "target_file", "-")
        record.target_line = getattr(record, "target_line", "-")
        return super().format(record)

class CallerAwareLogger(logging.Logger):
    """
    Logger that reports the first frame outside logging + abstract utilities.
    """

    def findCaller(self, stack_info=False, stacklevel=1):
        import inspect

        frame = inspect.currentframe().f_back  # ← critical fix

        while frame:
            code = frame.f_code
            filename = code.co_filename

            if not (
                "logging" in filename
                or "log_utils" in filename
                or "abstract_utilities" in filename
            ):
                return (
                    filename,
                    frame.f_lineno,
                    code.co_name,
                    None,
                )

            frame = frame.f_back

        return super().findCaller(stack_info, stacklevel)

# ─────────────────────────────────────────────────────────────
# Stack-aware logger
# ─────────────────────────────────────────────────────────────

class StackAwareLogger(logging.Logger):
    """
    Logger that automatically skips itself when reporting
    filename / line number.
    """

    _STACKLEVEL = 2

    def debug(self, msg, *args, **kwargs):
        kwargs.setdefault("stacklevel", self._STACKLEVEL)
        super().debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        kwargs.setdefault("stacklevel", self._STACKLEVEL)
        super().info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        kwargs.setdefault("stacklevel", self._STACKLEVEL)
        super().warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        kwargs.setdefault("stacklevel", self._STACKLEVEL)
        super().error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        kwargs.setdefault("stacklevel", self._STACKLEVEL)
        super().exception(msg, *args, **kwargs)


# Ensure our logger class is used globally (safe once)
logging.setLoggerClass(StackAwareLogger)


# ─────────────────────────────────────────────────────────────
# Log directory resolution
# ─────────────────────────────────────────────────────────────

def _resolve_log_root() -> Path:
    venv = os.getenv("VIRTUAL_ENV") or os.getenv("CONDA_PREFIX")
    if venv:
        p = Path(venv) / ".logs" / PACKAGE_NAME
        p.mkdir(parents=True, exist_ok=True)
        return p

    home = Path.home() / ".cache" / PACKAGE_NAME / "logs"
    try:
        home.mkdir(parents=True, exist_ok=True)
        return home
    except PermissionError:
        pass

    try:
        syslog = Path("/var/log") / PACKAGE_NAME
        syslog.mkdir(parents=True, exist_ok=True)
        return syslog
    except PermissionError:
        fallback = Path("/tmp") / PACKAGE_NAME / "logs"
        fallback.mkdir(parents=True, exist_ok=True)
        return fallback


LOG_ROOT = _resolve_log_root()


# ─────────────────────────────────────────────────────────────
# Public API
# ─────────────────────────────────────────────────────────────

def get_logFile(
    name: str,
    *,
    level: int = logging.INFO,
    console: bool = True,
    max_bytes: int = 5 * 1024 * 1024,
    backup_count: int = 5,
) -> logging.Logger:

    logger = logging.getLogger(name)

    if logger.handlers:
        return logger

    logger.setLevel(level)

    formatter = SafeFormatter(LOG_FORMAT, DATE_FORMAT)

    try:
        file_handler = RotatingFileHandler(
            LOG_ROOT / f"{name}.log",
            maxBytes=max_bytes,
            backupCount=backup_count,
            encoding="utf-8",
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    except PermissionError:
        logger.addHandler(logging.NullHandler())

    if console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    logger.propagate = False
    return logger




from .predicate_utils import *
def _get_default_modular(value, default, add=False, typ=set):
    """Merge user and default values intelligently."""
    if value == None:
        value = add
    if value in [True]:
        return default
    if value is False:
        return value
    if add:
        return combine_params(value,default,typ=None)

    return typ(value)

# -------------------------
# Default derivation logic
# -------------------------
def _get_default_modular(value, default, add=None, typ=set):
    """Merge user and default values intelligently."""
    add = add or False
    if value == None:
        value = add
    if value in [True]:
        return default
    if value is False:
        return value
    if add:
        return combine_params(value,default,typ=None)
    return typ(value)
def make_allowed_predicate(cfg: ScanConfig=None,**kwargs) -> Callable[[str], bool]:
    """
    Build a predicate that returns True if a given path is considered allowed
    under the given ScanConfig. Applies allowed_* and exclude_* logic symmetrically.
    """
    cfg=cfg or define_defaults(**kwargs)
    def allowed(path: str=None,p=None) -> bool:
        p = p or Path(path)
        name = p.name.lower()
        path_str = str(p).lower()

        # --------------------
        # A) directory filters
        # --------------------
        if cfg.exclude_dirs:
            for dpat in cfg.exclude_dirs:
                dpat_l = dpat.lower()
                if dpat_l in path_str or fnmatch.fnmatch(name, dpat_l):
                    if p.is_dir() or dpat_l in path_str:
                        return False

        if cfg.allowed_dirs and cfg.allowed_dirs != ["*"]:
            # must be in at least one allowed dir
            if not any(
                fnmatch.fnmatch(path_str, f"*{dpat.lower()}*") for dpat in cfg.allowed_dirs
            ):
                return False

        # --------------------
        # B) pattern filters
        # --------------------
        if cfg.allowed_patterns and cfg.allowed_patterns != ["*"]:
            if not any(fnmatch.fnmatch(name, pat.lower()) for pat in cfg.allowed_patterns):
                return False

        if cfg.exclude_patterns:
            for pat in cfg.exclude_patterns:
                if fnmatch.fnmatch(name, pat.lower()):
                    return False

        # --------------------
        # C) extension filters
        # --------------------
        if p.is_file():
            ext = p.suffix.lower()
            if cfg.allowed_exts and ext not in cfg.allowed_exts:
                return False
            if cfg.exclude_exts and ext in cfg.exclude_exts:
                return False

        # --------------------
        # D) type filters (optional)
        # --------------------
        if cfg.allowed_types and cfg.allowed_types != {"*"}:
            if not any(t in path_str for t in cfg.allowed_types):
                return False
        if cfg.exclude_types and cfg.exclude_types != {"*"}:
            if any(t in path_str for t in cfg.exclude_types):
                return False

        return True

    return allowed

def derive_all_defaults(**kwargs):
    kwargs = get_safe_canonical_kwargs(**kwargs)
    add = kwargs.get("add",False)
    nu_defaults = {}
    for key,values in DEFAULT_CANONICAL_MAP.items():
        default = values.get("default")
        typ = values.get("type")
        key_value = kwargs.get(key)
        if key in DEFAULT_ALLOWED_EXCLUDE_MAP:

            if key.endswith('exts'):
                input_value = ensure_exts(key_value)
            if key.endswith('patterns'):
                input_value = ensure_patterns(key_value)
            else:
                input_value = normalize_listlike(key_value, typ)
            nu_defaults[key] = _get_default_modular(input_value, default, add, typ)
        else:
            value = default if key_value is None else key_value
            if typ == list:
                value = make_list(value,commaparse=False)
            elif typ == bool:
                value = bool(value)
            nu_defaults[key] = value
   
    return nu_defaults
# -------------------------
# Default derivation logic
# -------------------------
def derive_file_defaults(**kwargs):
    kwargs = derive_all_defaults(**kwargs)
    add = kwargs.get("add",True)
    nu_defaults = {}
    for key,values in DEFAULT_ALLOWED_EXCLUDE_MAP.items():
        default = values.get("default")
        typ = values.get("type")
        key_value = kwargs.get(key)
        if key.endswith('exts'):
            input_value = ensure_exts(key_value)
        if key.endswith('patterns'):
            input_value = ensure_patterns(key_value)
        else:
            input_value = normalize_listlike(key_value, typ)
        nu_defaults[key] = _get_default_modular(input_value, default, add, typ)
    return nu_defaults

def define_defaults(**kwargs):
    defaults = derive_file_defaults(**kwargs)
    return ScanConfig(**defaults)

def get_file_filters(*args,**kwargs):
    directories = ensure_directories(*args,**kwargs)
    recursive = kwargs.get('recursive',True)
    include_files = kwargs.get('include_files',True)
    cfg = define_defaults(**kwargs)
    allowed = kwargs.get("allowed") or make_allowed_predicate(cfg)
    return directories,cfg,allowed,include_files,recursive
